package ru.dbotthepony.kstarbound.defs.tile

import com.google.gson.GsonBuilder
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.io.KConcreteTypeAdapter
import ru.dbotthepony.kstarbound.world.ITileGetter
import ru.dbotthepony.kvector.vector.nint.Vector2i
import java.util.concurrent.ConcurrentHashMap

data class RenderPiece(
	val texture: String? = null,
	val textureSize: Vector2i,
	val texturePosition: Vector2i,
	val colorStride: Vector2i? = null,
	val variantStride: Vector2i? = null,
) {
	companion object {
		val ADAPTER = KConcreteTypeAdapter.Builder(RenderPiece::class)
			.plain(
				RenderPiece::texture,
				RenderPiece::textureSize,
				RenderPiece::texturePosition,
				RenderPiece::colorStride,
				RenderPiece::variantStride,
			)
			.build()
	}
}

data class RenderRule(
	val type: String,
	val matchHue: Boolean = false,
	val inverse: Boolean = false,
) {
	private fun doTest(getter: ITileGetter, thisRef: TileDefinition, thisPos: Vector2i, offsetPos: Vector2i): Boolean {
		return when (type) {
			"EqualsSelf" -> getter[thisPos + offsetPos]?.def == thisRef
			"Connects" -> getter[thisPos + offsetPos] != null

			else -> false
		}
	}

	fun test(getter: ITileGetter, thisRef: TileDefinition, thisPos: Vector2i, offsetPos: Vector2i): Boolean {
		if (inverse) {
			return !doTest(getter, thisRef, thisPos, offsetPos)
		}

		return doTest(getter, thisRef, thisPos, offsetPos)
	}

	companion object {
		val ADAPTER = KConcreteTypeAdapter.Builder(RenderRule::class)
			.plain(
				RenderRule::type,
				RenderRule::matchHue,
				RenderRule::inverse,
			)
			.build()
	}
}

data class RenderRuleList(
	val entries: List<RenderRule>
) {
	companion object {
		val ADAPTER = KConcreteTypeAdapter.Builder(RenderRuleList::class)
			.list(RenderRuleList::entries, RenderRule::class)
			.build()
	}
}

data class RenderMatch(
	val pieces: List<Piece>? = null,
	val matchAllPoints: List<Matcher>? = null,
	val subMatches: List<RenderMatch>? = null,
	val haltOnMatch: Boolean = false
) {
	data class Piece(
		val name: String,
		val point: Vector2i
	)

	data class Matcher(
		val point: Vector2i,
		val ruleName: String
	)

	companion object {
		val ADAPTER = KConcreteTypeAdapter.Builder(RenderMatch::class)
			.list(RenderMatch::pieces, Piece::class)
			.list(RenderMatch::matchAllPoints, Matcher::class)
			.list(RenderMatch::subMatches, RenderMatch::class)
			.plain(RenderMatch::haltOnMatch)
			.build()

		val PIECE_ADAPTER = KConcreteTypeAdapter.Builder(Piece::class)
			.plain(Piece::name)
			.plain(Piece::point)
			.build(asList = true)

		val MATCHER_ADAPTER = KConcreteTypeAdapter.Builder(Matcher::class)
			.plain(Matcher::point)
			.plain(Matcher::ruleName)
			.build(asList = true)
	}
}

data class RenderMatchList(
	val name: String,
	val list: List<RenderMatch>
) {
	companion object {
		val ADAPTER = KConcreteTypeAdapter.Builder(RenderMatchList::class)
			.plain(RenderMatchList::name)
			.list(RenderMatchList::list, RenderMatch::class)
			.build(asList = true)
	}
}

data class RenderTemplate(
	val pieces: Map<String, RenderPiece>,
	val representativePiece: String,
	val matches: List<RenderMatchList>
) {
	companion object {
		val ADAPTER = KConcreteTypeAdapter.Builder(RenderTemplate::class)
			.map(RenderTemplate::pieces, RenderPiece::class.java)
			.plain(RenderTemplate::representativePiece)
			.list(RenderTemplate::matches, RenderMatchList::class.java)
			.build()

		fun registerGson(gsonBuilder: GsonBuilder) {
			RenderPiece.ADAPTER.register(gsonBuilder)
			RenderRule.ADAPTER.register(gsonBuilder)
			RenderRuleList.ADAPTER.register(gsonBuilder)
			RenderMatch.ADAPTER.register(gsonBuilder)
			RenderMatch.PIECE_ADAPTER.register(gsonBuilder)
			RenderMatch.MATCHER_ADAPTER.register(gsonBuilder)
			RenderMatchList.ADAPTER.register(gsonBuilder)

			ADAPTER.register(gsonBuilder)
		}

		private val cache = ConcurrentHashMap<String, RenderTemplate>()
		private val _readingFolder = ThreadLocal<String>()

		/**
		 * Случит переменной для указания из какой папки происходит чтение в данном потоке
		 */
		var readingFolder: String?
			get() = _readingFolder.get()
			set(value) { _readingFolder.set(value) }

		val CACHE = object : TypeAdapter<RenderTemplate>() {
			override fun write(out: JsonWriter, value: RenderTemplate) {
				ADAPTER.write(out, value)
			}

			override fun read(reader: JsonReader): RenderTemplate {
				if (reader.peek() != JsonToken.STRING) {
					throw JsonSyntaxException("Expected string as input for render template cache retriever")
				}

				var path = reader.nextString()

				if (path[0] != '/') {
					// относительный путь

					val readingFolder = readingFolder ?: throw NullPointerException("Currently read folder is not specified")
					path = "$readingFolder/$path"
				}

				return cache.computeIfAbsent(path) {
					return@computeIfAbsent Starbound.gson.fromJson(Starbound.getReader(it), RenderTemplate::class.java)
				}
			}
		}
	}
}
